/***************************************************************************
 *	Copyright (C) 2005 by Karye												*
 *	karye@users.sourceforge.net												*
 *																			*
 *	This program is free software; you can redistribute it and/or modify	*
 *	it under the terms of the GNU General Public License as published by	*
 *	the Free Software Foundation; either version 2 of the License, or		*
 *	(at your option) any later version.										*
 *																			*
 *	This program is distributed in the hope that it will be useful,			*
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of			*
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the			*
 *	GNU General Public License for more details.							*
 *																			*
 *	You should have received a copy of the GNU General Public License		*
 *	along with this program; if not, write to the							*
 *	Free Software Foundation, Inc.,											*
 *	59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.				*
 ***************************************************************************/

#include <qdebug.h>       // for QDebug
#include <qdir.h>         // for QDir, operator|, QDir::Name, QDir::NoSymLinks
#include <qglobal.h>      // for qWarning, foreach, qCritical
#include <qlist.h>        // for QList<>::iterator, QList<>::Iterator, QList
#include <qmap.h>         // for QMap, QMap<>::iterator
#include <qobjectdefs.h>  // for Q_EMIT
#include <qstring.h>      // for QString, operator+, operator<
#include <qstringlist.h>  // for QStringList

#include "cacheportagejob.h"
#include "common.h"       // for KurooDBSingleton, SignalistSingleton, DEBUG...
#include "global.h"       // for parsePackage
#include "portagedb.h"    // for KurooDB, DbConnection
#include "progressjob.h"  // for ProgressJob
#include "settings.h"     // for KurooConfig
#include "signalist.h"    // for Signalist

namespace ThreadWeaver {
class Thread;
}  // namespace ThreadWeaver

/**
 * @class CachePortageJob
 * @short Thread to cache package information from the Portage directory to speed up portage refreshing.
 *
 * Portage cache is scanned for package sizes, and stored in portage cache map and in the database.
 */
CachePortageJob::CachePortageJob()
    : 
	m_db( KurooDBSingleton::Instance()->getStaticDbConnection() )
{
}

CachePortageJob::~CachePortageJob()
{
	KurooDBSingleton::Instance()->returnStaticDbConnection( m_db );
}

/**
 * Scan for package size found in digest files.
 */
void CachePortageJob::run( ThreadWeaver::JobPointer self, ThreadWeaver::Thread*  /*thread*/)
{
	DEBUG_LINE_INFO;
	if ( !m_db->isConnected() ) {
		qCritical() << "Creating cache. Can not connect to database";
		return;// false;
	}

	int count( 0 );
	QMap <QString, QString> mapCache;
	QDir dCategory, dPackage;
	dCategory.setFilter( QDir::Dirs | QDir::NoSymLinks );
	dCategory.setSorting( QDir::Name );

	// Get a count of total packages for proper progress
	QString packageCount = KurooDBSingleton::Instance()->singleQuery( QStringLiteral("SELECT data FROM dbInfo WHERE meta = 'packageCount' LIMIT 1;"), m_db );
    if ( packageCount == u'0' )
		setTotalSteps( 35000 );
	else
        setTotalSteps( packageCount.toInt() );
    //TODO: setStatus( "CachePortage", i18n("Collecting package information...") );

	//TODO: kuroo was already built to handle overlays?
	// Get list of categories in Portage and Overlays
	QStringList pathList = KurooConfig::dirPortage().split(u' ');
	const QStringList pathOverlays = KurooConfig::dirPortageOverlay().split(u' ');
	for ( const QString& path : pathOverlays ) {
		pathList += path;
	}

	// Scan Portage cache
	for (auto & itPath : pathList) {
		if ( !dCategory.cd( itPath ) ) {
			qWarning() << "Creating cache. Can not access " << itPath;
			continue;
		}

		QStringList categoryList = dCategory.entryList();
		QStringList::Iterator itCategoryEnd = categoryList.end();
		for ( QStringList::Iterator itCategory = categoryList.begin(); itCategory != itCategoryEnd; ++itCategory ) {
			//TODO: Filter out distfiles, eclass, licences, metadata, packages, profiles, scripts
			if ( *itCategory == u'.' || *itCategory == QStringLiteral("..") )
				continue;

            // Abort the scan
			if ( isAborted() ) {
				qWarning() << "Creating cache. Aborted!";
				//TODO: setStatus( "CachePortage", i18n("Caching aborted.") );
				Q_EMIT failed( self );
				SignalistSingleton::Instance()->cachePortageComplete();
				return;
            }

			// Get list of packages in this category
			dPackage.setFilter( QDir::Files | QDir::NoSymLinks );
			dPackage.setSorting( QDir::Name );
			if ( dPackage.cd( itPath + QStringLiteral("/metadata/md5-cache/") + *itCategory ) ) {
				QStringList packageList = dPackage.entryList();
				QStringList::Iterator itPackageEnd = packageList.end();
				for ( QStringList::Iterator itPackage = packageList.begin(); itPackage != itPackageEnd; ++itPackage ) {

					if ( *itPackage == u'.' || *itPackage == QStringLiteral("..") )
						continue;

                    // Abort the scan
					if ( isAborted() ) {
						qWarning() << "Creating cache. Aborted!";
						//TODO: setStatus( "CachePortage", i18n("Caching aborted.") );
						Q_EMIT failed( self );
						SignalistSingleton::Instance()->cachePortageComplete();
						return;
                    }
					QString package = *itCategory + u'/' + *itPackage;

                    QStringList parts = parsePackage( *itPackage );
					if ( !parts.isEmpty() ) {
						QString packageName = parts[1];

//TODO: /files/digest-* doesn't seem to exist anymore, maybe we could read from 'Manifest' instead?
						// Get package size
						mapCache.insert( package, QStringLiteral("12345") );
						/*QString path = *itPath + "/" + *itCategory + "/" + packageName + "/files/digest-" + *itPackage;
						QFile file( path );
						if ( file.open( QIODevice::ReadOnly ) ) {
							std::ifstream in( path );
							std::string word;
							while ( in >> word );
							mapCache.insert( package, word );
							file.close();
						}
						else
							qWarning() << "Creating cache. Reading: " << path;*/
					}
					else
						qWarning() << "Creating cache. Can not parse: " << *itPackage;

					// Post scan count progress
                    if ( (++count % 100) == 0 )
                        setProgress( count );
				}
			}
			else
				qWarning() << "Creating cache. Can not access " << itPath << "/metadata/cache/" << *itCategory;

		}
	}
	KurooDBSingleton::Instance()->query( QStringLiteral("UPDATE dbInfo SET data = '%1' WHERE meta = 'packageCount';")
	                                     .arg( count ), m_db );

	// Store cache in DB
	KurooDBSingleton::Instance()->query( QStringLiteral("DELETE FROM cache;"), m_db );
	KurooDBSingleton::Instance()->query( QStringLiteral("BEGIN TRANSACTION;"), m_db );
	QMap< QString, QString >::iterator itMapEnd = mapCache.end();
	for ( QMap< QString, QString >::iterator itMap = mapCache.begin(); itMap != itMapEnd; ++itMap )
		KurooDBSingleton::Instance()->insert( QStringLiteral("INSERT INTO cache (package, size) VALUES ('%1', '%2');").
		                                      arg( itMap.key(), itMap.value() ), m_db );

	KurooDBSingleton::Instance()->query(QStringLiteral("COMMIT TRANSACTION;"), m_db );

    //TODO: setStatus( "CachePortage", i18n("Done.") );
    setProgress( 0 );
	DEBUG_LINE_INFO;
// 	return true;
// }
//
// void CachePortageJob::completeJob()
// {
	SignalistSingleton::Instance()->cachePortageComplete();
}

